//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Extract.cc --- implementation of Extractor class, as well as the
//                     Product base class.
//
//	CHANGE HISTORY:
//		UL, 20/02/98: 1.0.0: - skeleton (to be filled with flesh by DH)
//              DAH, 25/03/98: 1.0.0: - first real version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method
//
//	AUTHOR:
//              D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description of SAS task `evselect'
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/evselect/src/Extractor.cc,v 1.5 2012/03/29 17:07:06 aibarra Exp $
//

#include <sstream>
#include <cstring>

#include "param.h"
#include "Dal.h"
#include "dal_utilities.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>
#include "DataSubspace.h"
#include "xmmexposure.h"
#include "Extractor.h"

using namespace std;

void Extractor::doIt(Table *table, bool destructive, const std::string &flagColumn,
		     int flagBit, bool writeDss)
{
  if( booleanParameter("withimageset") ){
    Image image(stringParameter("imageset"), table, destructive, flagColumn, flagBit);
    image.make(writeDss);
  }

  if( booleanParameter("withspectrumset") ){
    Spec spec(stringParameter("spectrumset"), table, destructive, flagColumn, flagBit);
    spec.make(writeDss);
  }

  if( booleanParameter("withrateset") ){
    Rate rate(stringParameter("rateset"), table, destructive, flagColumn, flagBit);
    rate.make(writeDss);
  }

  if( booleanParameter("withhistogramset") ){
    Histo histo(stringParameter("histogramset"), table, destructive, flagColumn, flagBit);
    histo.make(writeDss);
  }
} // Extractor::doIt



Product::Product(const std::string &name, Table *table,
		 bool destructive, const std::string &flagColumn, int flagBit)
  :  _flagCol(flagColumn) {
  _intable = table;
  _destruct = destructive;

  // Make temporary table with the same size as the input table

  DataSet *tempSet = dataSetServer->open(tempnam(0,0), DataSet::Temp);
  _tempTable = tempSet->addTable("TempTable", _intable->rows());

  unsigned nDigits = 0;

  if(! _destruct){
    switch (table->column(_flagCol)->dataType()) {
    case Dal::Int8 : nDigits = 8; break;
    case Dal::Int16: nDigits = 16; break;
    case Dal::Int32: nDigits = 32; break;
    default: errstr<<"Filter::bitPosition: Flag column must be of"
	       "integral type"<<Sas::fatal("NoIntFlag"); break;
    }

    if ( flagBit == -1 || static_cast<unsigned>(flagBit) > nDigits-1){
      _flagBit = 1 << (nDigits-1);
    }else{
      _flagBit = 1<<flagBit;
    }
  }

  _dataSet = dataSetServer->open(name, DataSet::Create);

   _globalAttribs.push_back("TELESCOP");
   _globalAttribs.push_back("INSTRUME");
   _globalAttribs.push_back("SLCTEXPR");
   _globalAttribs.push_back("FILTER");

   std::string atrribsToCopy = "attributestocopy";
   for(unsigned ipos=0; ipos < (unsigned)parameterCount(atrribsToCopy); ipos++)
     _globalAttribs.push_back(stringParameter(atrribsToCopy,ipos));
}


Product::~Product(){

  // Copy any additional extensions requested by the user

  std::string slocks = "blockstocopy";

  DataSet * eventset = _intable->parent();

  for(unsigned ipos=0; ipos < (unsigned)parameterCount(slocks); ipos++)
    _dataSet->add( eventset->block(stringParameter(slocks,ipos)) );

  dataSetServer->close(_dataSet);
  dataSetServer->close(_tempTable->parent());
}


void Product::_copyGlobalAttribs(Block *out){

  // Copy all attributes from the input dataset to the ouput dataset

  copyAttributes(*out->parent(), *_intable->parent());

  // Copy the listed attributes to the Block

  for(unsigned int i=0; i<_intable->comments(); i++){
    out->addHistory( _intable->comment(i) );
  }

  for(unsigned int i=0; i<_intable->historys(); i++){
    out->addComment( _intable->history(i) );
  }

  // Is the block in the primary array?
  bool primBlock = out->name()=="PRIMARY";

  std::vector<std::string>::const_iterator iatt;
  for(iatt=_globalAttribs.begin(); iatt!=_globalAttribs.end(); iatt++){

    if( _intable->parent()->hasAttribute(*iatt) ){
      if(!primBlock) out->addAttribute( _intable->parent()->attribute(*iatt) );

    }else if( _intable->hasAttribute(*iatt) ){
      out->addAttribute( _intable->attribute(*iatt) );

    }else{
      errstr << "The Attribute " << (*iatt)
	     << " does not exist in the input event list. No value was copied."
	     << Sas::warning("NoAttrib");
    }
  }
}

void Product::getRange(const Column * col, real64 & min, real64 & max){

  // The following macro sillyness is necessary due to lack of auto-type conversion in the Dal

  Column *flagDatCol;
  ColumnData *cFlagData = 0;
  int32 *flagData = 0;

  if(! _destruct){
    flagDatCol = reType(col->parent()->column(_flagCol), Dal::Int32);
    cFlagData = flagDatCol->data();
    flagData = cFlagData->int32Data();
  }


  #define CODE(type)                                          \
  {                                                           \
    type * Data = Cdata->type##Data();                        \
                                                              \
    bool first = true;                                        \
    unsigned long mRows = col->parent()->rows();              \
    for(unsigned long i=0; i<mRows; i++){                     \
      if( ! _destruct && ((flagData[i] & _flagBit) == 0) ) continue; \
      if( first ){                                            \
	first = false;                                        \
	min = max = Data[i];                                  \
      }                                                       \
      if( Data[i] < min ) min = Data[i];                      \
      if( Data[i] > max ) max = Data[i];                      \
    }                                                         \
                                                              \
    break;                                                    \
  }                                                           \

  ColumnData *Cdata = col->data();

  switch(col->dataType()){

  case Dal::Int32  : CODE(int32)

  case Dal::Int16  : CODE(int16)

  case Dal::Int8   : CODE(int8)

  case Dal::Real32 : CODE(real32)

  case Dal::Real64 : CODE(real64)

  default :
    errstr << "getRange: column is not of a supported data type." << Sas::error("RangeType");
  }

  #undef CODE

  delete Cdata;
  if(! _destruct) delete cFlagData;
}


void Product::GetLimits(const Column *col, std::string minparam, std::string maxparam, bool useparam,
			std::vector<real64> &limits){

  // Get the min and max values for the column

  real64 minRange=0, maxRange=0;
  bool useTL = !booleanParameter("ignorelegallimits");

  // Do lower limit first.

  if(useparam){

    // If parameter is set, this is what min limit should be
    limits[0] = realParameter(minparam);

    // If this is below allowed range, adjust and warn user.
    if( useTL && col->hasAttribute("TLMIN") && limits[0] < col->attribute("TLMIN")->asReal() ){
      limits[0] = col->attribute("TLMIN")->asReal();
      errstr <<"Value of parameter "<<minparam<<" outside allowed range for column "
	     <<col->name()<<"; Setting to minimum value."<<Sas::warning("MinParamRange");
    }

  }else if(useTL && col->hasAttribute("TLMIN")){

    // No parameter, so try setting to TLMIN
    limits[0] = col->attribute("TLMIN")->asReal();

    // Check if data is outside allowed range, warn user.
    //    if( limits[0] > minRange ){
    //      errstr <<"Data values found outside of allowed range for column "<<col->name()
    //	     <<"; will be ignored."<<Sas::warning("MinDataRange");
    //    }

  }else{

    // No guidance from user, so just set limit to lower range of data.
    getRange(col, minRange, maxRange);
    limits[0] =  minRange;
  }


// Now do upper limit.

  if(useparam){

    // If parameter is set, this is what max limit should be
    limits[1] = realParameter(maxparam);

    // If this is above allowed range, adjust and warn user.
    if( useTL && col->hasAttribute("TLMAX") && limits[1] > col->attribute("TLMAX")->asReal() ){
      limits[1] = col->attribute("TLMAX")->asReal();
      errstr <<"Value of parameter "<<maxparam<<" outside allowed range for column "
	     <<col->name()<<"; Setting to maximum value."<<Sas::warning("MaxParamRange");
    }

  }else if( useTL && col->hasAttribute("TLMAX") ){

    // No parameter, so try setting to TLMAX
    limits[1] = col->attribute("TLMAX")->asReal();

    // Check if data is outside allowed range, warn user.
    //    if( limits[1] < maxRange ){
    //      errstr <<"Data values found outside of allowed range for column "<<col->name()
    //	     <<"; will be ignored."<<Sas::warning("MaxDataRange");
    //    }

  }else{

    // No guidance from user, so just set limit to upper range of data.
    if(minRange==0 && maxRange==0) getRange(col, minRange, maxRange);
    limits[1] =  maxRange;
  }

  // Done.
}

// AI: New method to calculate the Observation Start Time and Stop Time using the
// TSTART and TSTOP keywords. SAS-SCR-238
void Product::GetObservationLimits(const Column *col, std::string minparam, std::string maxparam, bool useparam,
			std::vector<real64> &limits){

  // Get the min and max values for the column
  real64 minRange=0, maxRange=0;
  bool useTL = !booleanParameter("ignorelegallimits");

  // Do lower limit first.

  if(useparam)
  {
	  // If parameter is set, this is what min limit should be
	  limits[0] = realParameter(minparam);

	  // If this is below allowed range, adjust and warn user.
	  if( useTL && col->hasAttribute("TLMIN") && limits[0] < col->attribute("TLMIN")->asReal() )
	  {
		  limits[0] = col->attribute("TLMIN")->asReal();
		  errstr <<"Value of parameter "<<minparam<<" outside allowed range for column "
		  <<col->name()<<"; Setting to minimum value."<<Sas::warning("MinParamRange");
	  }
  }
  else if(useTL &&  _intable->hasAttribute("TSTART") )
  {
	  // No parameter. First check if the TSTART keyword exists. If not
	  // if not, takes the first value of the column.
	  limits[0] = _intable->attribute("TSTART")->asReal();
  }
  else
  {
    // No guidance from user, so just set limit to lower range of data.
    getRange(col, minRange, maxRange);
    limits[0] =  minRange;
  }


// Now do upper limit.

  if(useparam)
  {
	  // If parameter is set, this is what max limit should be
	  limits[1] = realParameter(maxparam);

	  // If this is above allowed range, adjust and warn user.
	  if( useTL && col->hasAttribute("TLMAX") && limits[1] > col->attribute("TLMAX")->asReal() )
	  {
		  limits[1] = col->attribute("TLMAX")->asReal();
		  errstr <<"Value of parameter "<<maxparam<<" outside allowed range for column "
		  <<col->name()<<"; Setting to maximum value."<<Sas::warning("MaxParamRange");
	  }
  }
  else if( useTL && _intable->hasAttribute("TSTOP") )
  {
	  // No parameter, so try setting to TLMAX
	  limits[1] = _intable->attribute("TSTOP")->asReal();;
  }
  else
  {
    // No guidance from user, so just set limit to upper range of data.
    if(minRange==0 && maxRange==0) getRange(col, minRange, maxRange);
    limits[1] =  maxRange;
  }

  // Done.
}


Column * Product::reType(const Column *oldCol, Dal::DataType newType){

  // if( newType == oldCol->dataType() ) return oldCol;

  std::string newName;
  newName = "jnk_" + oldCol->name();
  while( _tempTable->hasColumn(newName) ) newName=newName+"A";

  #define CODE(otype, ntype)                                                         \
  {                                                                                  \
    ColumnData * cNewData = _tempTable->addColumn(newName, newType)->data();         \
    ColumnData * cOldData = oldCol->data();                                          \
    ntype * newData = cNewData->ntype##Data();                                       \
    otype * oldData = cOldData->otype##Data();                                       \
    for(unsigned long i = 0; i < _intable->rows(); i++){                             \
      newData[i] = static_cast<ntype>(oldData[i]);                                   \
    }                                                                                \
    delete cNewData;                                                                 \
    delete cOldData;                                                                 \
  }                                                                                  \
  break;                                                                             \

  switch(oldCol->dataType()){

  case Dal::Int8 : {
    switch(newType){
    case Dal::Int8   : CODE(int8, int8)
    case Dal::Int16  : CODE(int8, int16)
    case Dal::Int32  : CODE(int8, int32)
    case Dal::Real32 : CODE(int8, real32)
    case Dal::Real64 : CODE(int8, real64)
    default :
      errstr << "Programming error: Column type conversion not supported." << Sas::error("ColType");
    }
  }
  break;

  case Dal::Int16 : {
    switch(newType){
    case Dal::Int8   : CODE(int16, int8)
    case Dal::Int16  : CODE(int16, int16)
    case Dal::Int32  : CODE(int16, int32)
    case Dal::Real32 : CODE(int16, real32)
    case Dal::Real64 : CODE(int16, real64)
    default :
      errstr << "Programming error: Col type conversion not supported." << Sas::error("ColType");
    }
  }
  break;

  case Dal::Int32 : {
    switch(newType){
    case Dal::Int8   : CODE(int32, int8)
    case Dal::Int16  : CODE(int32, int16)
    case Dal::Int32  : CODE(int32, int32)
    case Dal::Real32 : CODE(int32, real32)
    case Dal::Real64 : CODE(int32, real64)
    default :
      errstr << "Programming error: Col type conversion not supported." << Sas::error("ColType");
    }
  }
  break;

  case Dal::Real32 : {
    switch(newType){
    case Dal::Int8   : CODE(real32, int8)
    case Dal::Int16  : CODE(real32, int16)
    case Dal::Int32  : CODE(real32, int32)
    case Dal::Real32 : CODE(real32, real32)
    case Dal::Real64 : CODE(real32, real64)
    default :
      errstr << "Programming error: Col type conversion not supported." << Sas::error("ColType");
    }
  }
  break;

  case Dal::Real64 : {
    switch(newType){
    case Dal::Int8   : CODE(real64, int8)
    case Dal::Int16  : CODE(real64, int16)
    case Dal::Int32  : CODE(real64, int32)
    case Dal::Real32 : CODE(real64, real32)
    case Dal::Real64 : CODE(real64, real64)
    default :
      errstr << "Programming error: Col type conversion not supported." << Sas::error("ColType");
    }
  }
  break;

  default :
    errstr << "Column " <<oldCol->name()<< " is not of a supported data type." << Sas::error("ColType");
  }

  #undef CODE

  return _tempTable->column(newName);
}


void Product::_getExposure(Block *productBlock){

  if( !_intable->parent()->hasAttribute("TELESCOP") ||
      _intable->parent()->attribute("TELESCOP")->asString() != "XMM" ){
    errstr << "Not XMM data.  Cannot calculate EXPOSURE keyword for output products."
	   << Sas::warning("NotXmmProduct");
    return;
  }

  std::string instrument = _intable->parent()->attribute("INSTRUME")->asString();

  if(instrument=="RGS1" || instrument=="RGS2"){

    if( !_intable->hasAttribute("EXPOSURE") ){
      errstr << "No EXPOSURE keyword in RGS event list.  "
	     <<	"Cannot write exposure value to the product file."
	     << Sas::warning("NoRGSExposure");
      return;
    }

    productBlock->addAttribute(_intable->attribute("EXPOSURE"));

    // Find 'first' region extension, and copy keywords to product file.

    DataSubspace dss(productBlock);
    bool spaceAttributes=false, orderAttributes=false;
    for(unsigned icomp=0; icomp<dss.components(); icomp++){
      DScomp * comp = dss.component(icomp);
      for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
	DSfilter *filter = comp->filter(ifilter);
	if(filter->type() == DSS::RegionT){
	  DSS::DSfilterComp xycomps = filter->comps();
	  std::string blockSpec = filter->blockSpec();
	  DataSet *set = dataSetServer->open(blockSpec, Dal::Read);
	  Table *table = set->table(blockSpec);

	  if( !orderAttributes && xycomps[0]=="BETA_CORR" && xycomps[1]=="PI" ){
	    if( !table->hasAttribute("ORDER") ) continue;
	    orderAttributes=true;
	    productBlock->addAttribute(table->attribute("ORDER"));
	  }

	  if( !spaceAttributes && xycomps[0]=="BETA_CORR" && xycomps[1]=="XDSP_CORR" ){
	    if( !table->hasAttribute("SOURCEID") ) continue;
	    if( !table->hasAttribute("RA_SRC") ) continue;
	    if( !table->hasAttribute("DEC_SRC") ) continue;
	    if( !table->hasAttribute("RA_REF") ) continue;
	    if( !table->hasAttribute("DEC_REF") ) continue;
	    if( !table->hasAttribute("APOS_REF") ) continue;
	    if( !table->hasAttribute("OFFAXIS1") ) continue;
	    if( !table->hasAttribute("OFFAXIS2") ) continue;
	    spaceAttributes=true;
	    productBlock->addAttribute(table->attribute("SOURCEID"));
	    productBlock->addAttribute(table->attribute("RA_SRC"));
	    productBlock->addAttribute(table->attribute("DEC_SRC"));
	    productBlock->addAttribute(table->attribute("RA_REF"));
	    productBlock->addAttribute(table->attribute("DEC_REF"));
	    productBlock->addAttribute(table->attribute("APOS_REF"));
	    productBlock->addAttribute(table->attribute("OFFAXIS1"));
	    productBlock->addAttribute(table->attribute("OFFAXIS2"));
	  }

	  if( spaceAttributes && orderAttributes ) break;
	}
      }
      if( spaceAttributes && orderAttributes ) break;
    }

    if( !spaceAttributes || !orderAttributes )
      errstr << "Could not find requested RGS attributes to copy to "
	     << "the product extension." << Sas::warning("RGSprodAttributes");

  }else if(instrument=="EMOS1" || instrument=="EMOS2" || instrument=="EPN"){

    std::vector<unsigned long> ccdCounts(17);

    if( !_intable->hasColumn("CCDNR") ){
      errstr << "Event list does not have a CCDNR column. "
	     << "Cannot calculate EXPOSURE keyword for EPIC products."
	     << Sas::warning("NoCCDcolumn");
      return;
    }

    ColumnData * ccdCdat = _intable->column("CCDNR")->data();
    int8 * ccdDat = ccdCdat->int8Data();

    Column *flagDatCol;
    ColumnData *cFlagData = 0;
    int32 *flagData = 0;
    if(! _destruct){
      flagDatCol = reType(_intable->column(_flagCol), Dal::Int32);
      cFlagData = flagDatCol->data();
      flagData = cFlagData->int32Data();
    }

    unsigned long nevents=0;
    unsigned long mRows = _intable->rows();

    for(unsigned long ievent=0; ievent<mRows; ievent++){
      if( ! _destruct && ((flagData[ievent] & _flagBit) == 0) ) continue;
      ccdCounts[ccdDat[ievent]]++;
      nevents++;
    }

    if(nevents==0){
      errstr << "Cannot calculate exposure when no events are selected. "
	     << "Cannot write EXPOSURE keyword for output products."
	     << Sas::warning("NoExpEmpty");
      return;
    }

    delete ccdCdat;
    if(! _destruct) delete cFlagData;

    real64 exposure=0;

    if( _intable->hasAttribute("LIVETIME") ){

      for(unsigned iccd=0; iccd<17; iccd++){
	if(ccdCounts[iccd]){
	  std::stringstream keyName;
	  keyName << "LIVETI";
	  keyName.width(2);
	  keyName.fill('0');
	  keyName << iccd;

	  if( !_intable->hasAttribute(keyName.str()) ) continue;

	  exposure += ccdCounts[iccd] / _intable->attribute(keyName.str())->asReal();
	}
      }

    }else{

      errstr << "No live time information available.  Cannot write "
	     << "EXPOSURE keyword for output products." << Sas::warning("NoProductExp");
      return;
    }

    *productBlock->addAttribute("EXPOSURE",
	      "Weighted live time of CCDs in the extraction region") = nevents/exposure;
  }

  return;
}


void Product::_doColGen(ExprColumnGenerator &colGen, const Table &inTable, Table &outTable){
    try {
      colGen.doIt(inTable, outTable);
    } catch (SelectException except) {
      errstr<< except.text() << Sas::error("ExprColGenError");
    }
}

