//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	spec.cc -- implementation of the Spec class
//
//	CHANGE HISTORY:
//              DAH, 25/03/98: 1.0.0: - first version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method
//
//	AUTHOR:
//              D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description of SAS task `evselect'
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/evselect/src/rate.cc,v 1.5 2013/09/19 16:30:46 aibarra Exp $

#include <iostream>
#include <sstream>
#include <cmath>
#include "param.h"
#include "Dal.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>
#include "Extractor.h"
#include "DataSubspace.h"
#include <iomanip>
void Rate::make(bool writeDss){

   _globalAttribs.push_back("TIMESYS");
   _globalAttribs.push_back("TIMEUNIT");
   _globalAttribs.push_back("TIMEREF");
   _globalAttribs.push_back("TASSIGN");
   _globalAttribs.push_back("CLOCKAPP");
   _globalAttribs.push_back("MJDREF");

   bool zcol = booleanParameter("withzcolumn");
   bool zerrcol = booleanParameter("withzerrorcolumn");
   bool ratecol = booleanParameter("makeratecolumn");

  // Get pointers to the needed columns

  Column *tDatCol, *zDatCol = 0, *zDatErrCol = 0;
  Column *tCol, *zCol, *zErrCol;

  std::string tcolumn = stringParameter("timecolumn");
  if( _intable->hasColumn(tcolumn) ){
    tCol = _intable->column(tcolumn);
    tDatCol = reType(tCol, Dal::Real64);
  }else{
    std::stringstream newName;
    newName << "T" << _intable;
    ExprColumnGenerator colGen(tcolumn, newName.str());
    _doColGen(colGen, *_intable, *_tempTable);
    tCol = tDatCol = _intable->column(tcolumn);
  }

  if(zcol){
    std::string zcolumn = stringParameter("zcolumn");
    if( _intable->hasColumn(zcolumn) ){
      zCol = _intable->column(zcolumn);
      zDatCol=reType(zCol, Dal::Real64);
    }else{
      std::stringstream newName;
      newName << "Z" << _intable;
      ExprColumnGenerator colGenZ(zcolumn, newName.str());
      _doColGen(colGenZ, *_intable, *_tempTable);
      zCol = zDatCol = _intable->column(zcolumn);
    }

    if(zerrcol){
      zcolumn = stringParameter("zerrorcolumn");
      if( _intable->hasColumn(zcolumn) ){
	zErrCol = _intable->column(zcolumn);
	zDatErrCol=reType(zErrCol, Dal::Real64);
      }else{
	std::stringstream newName;
	newName << "ZE" << _intable;
	ExprColumnGenerator colGenZ(zcolumn, newName.str());
	_doColGen(colGenZ, *_intable, *_tempTable);
	zErrCol = zDatErrCol = _intable->column(zcolumn);
      }
    }
  }

  bool tIntCol = tCol->dataType()==Dal::Int8 || tCol->dataType()==Dal::Int16 ||
                 tCol->dataType()==Dal::Int32;

  // Read in the needed parameters

  _binSize = realParameter("timebinsize");
  if( _binSize==0 )
    errstr << "Time bin size set to zero.  This does not make sense." <<Sas::error("RateZeroBinVal");


  if( tIntCol && floor(_binSize) != _binSize ){
    errstr << "Binning factor must be an integer for integer-valued columns."<<Sas::error("RateBinVal");
  }

  if( ( _intable->hasAttribute("FRTIME") && _intable->attribute("FRTIME")->asReal() > _binSize ) ||
      ( _intable->parent()->hasAttribute("FRTIME") && _intable->parent()->attribute("FRTIME")->asReal() > _binSize ) ){

    errstr << "Time bin size is less than the mean frame readout time." << Sas::warning("BinVsFrameTime");
  }

  // Determine the time range

  std::vector<real64> useRange(2);
  // AI: New method to calculate the observation limits, taking into account the
  // observation start time and end time. SAS-SCR-238
  //GetLimits(tCol, "timemin", "timemax", booleanParameter("withtimeranges"), useRange);
  GetObservationLimits(tCol, "timemin", "timemax", booleanParameter("withtimeranges"), useRange);

  if( tIntCol ){
    if( useRange[0] != floor(useRange[0]) || useRange[1] != floor(useRange[1]) ){
      errstr << "Time ranges must be integers for integer-valued columns."<<Sas::error("TimeRangeVal");
    }
    useRange[0] -= 0.5;
    useRange[1] += 0.5;
  }

  if( useRange[1] == useRange[0] )
    errstr << "The time column for extracting the lightcurve has "
	   << "the same value throughout, or the upper and lower limits "
	   << "for this column are equal. An empty lightcurve will "
	   << "be created." << Sas::warning("BadRateLimits");
  if( useRange[1] < useRange[0] )
    errstr << "Upper rate limit < lower limit.  Cannot continue."
	   << Sas::error("BadRateLimits");

  unsigned long nRows = (unsigned long)ceil((useRange[1]-useRange[0])/_binSize);

  // Now initialize the output light curve

  Table * outTable = _dataSet->addTable("RATE", nRows);

  _copyGlobalAttribs(outTable);

  // Write some keywords

  *outTable->addAttribute( "HDUCLASS",
			   "Format conforms to OGIP/GSFC conventions")
    = "OGIP";

  *outTable->addAttribute( "HDUCLAS1",
			    "File contains a time-serie")
    = "LIGHTCURVE";


  if (ratecol)
    {
      *outTable->addAttribute( "HDUCLAS2",
			       "File contains gross rates")
	= "TOTAL";

      *outTable->addAttribute( "HDUCLAS3",
			       "Light curve stored as rates")
	= "RATE";
    }
  else
    {
      *outTable->addAttribute( "HDUCLAS2",
			       "File contains gross counts")
	= "TOTAL";

      *outTable->addAttribute( "HDUCLAS3",
			       "Light curve stored as counts")
	= "COUNT";
    }
  *outTable->addAttribute( "HDUVERS1",
			   "Version of format")
    = "1.1.0";


  // Rate file specific keywords.

  *outTable->addAttribute( "TIMEDEL",
			   "Light curve bin size")
    = tIntCol ? (int32)_binSize : _binSize;
  *outTable->addAttribute( "TSTART",
			   "Light curve start time")
    = useRange[0];

  *outTable->addAttribute( "TSTOP",
			   "Light curve end time")
    = useRange[1];

  // Make the necessary columns.

  int32 *cntDat=0;
  real64 *rateDat=0, *rateErrDat=0;
  real64 *zData=0, *zErrData=0;
  ColumnData *cRateDat=0, *cErrRateDat=0, *cZDatCol=0, *cZDatErrCol=0, *cCntDat=0;

  if( zcol || ratecol ){
    cRateDat = outTable->addColumn("RATE", Dal::Real64, "", "count/s")->data();
    cErrRateDat = outTable->addColumn("ERROR", Dal::Real64, "", "count/s")->data();
    rateDat = cRateDat->real64Data();
    rateErrDat = cErrRateDat->real64Data();
    for(unsigned i=0; i<nRows; i++) rateDat[i] = rateErrDat[i] = 0;

    *outTable->addAttribute( "BACKAPP", "Background correction applied?") = false;
    *outTable->addAttribute( "DEADAPP", "Dead time correction applied?") = false;
    *outTable->addAttribute( "VIGNAPP", "Vignetting correction applied?") = false;
  }

  if(zcol){
    cZDatCol = zDatCol->data();
    zData = cZDatCol->real64Data();
    if(zerrcol){
      cZDatErrCol = zDatErrCol->data();
      zErrData = cZDatErrCol->real64Data();
    }
  }

  cCntDat = outTable->addColumn("COUNTS", Dal::Int32, "", "count")->data();
  cntDat = cCntDat->int32Data();
  for(unsigned i=0; i<nRows; i++) cntDat[i] = 0;


  real64 timeZero = useRange[0] + _binSize/2.;

  if( _intable->hasAttribute("TIMEZERO") ){
    timeZero += _intable->attribute("TIMEZERO")->asReal();
  }

  if( booleanParameter("maketimecolumn") ){
    ColumnData *cTimeDat= outTable->addColumn("TIME", Dal::Real64, "", "s")->data();
    real64 *timeDat = cTimeDat->real64Data();
    for(unsigned i=0; i<nRows; i++) timeDat[i] = timeZero + i*_binSize;

    delete cTimeDat;
  }else{
    *outTable->addAttribute( "TIMEZERO",
			     "Time of first light curve bin")
      = timeZero;
  }

  // Set up type converted columns

  Column *flagDatCol;

  ColumnData *cTDatCol = tDatCol->data(), *cFlagData = 0;
  int32 index, *flagData = 0;
  real64 *tData = cTDatCol->real64Data();

  if(! _destruct){
    flagDatCol = reType(_intable->column(_flagCol), Dal::Int32);
    cFlagData = flagDatCol->data();
    flagData = cFlagData->int32Data();
  }

  // Accumulate the columns

  unsigned long oRows = _intable->rows();
  for(unsigned long irow=0; irow<oRows; irow++){
    if( ! _destruct && ((flagData[irow] & _flagBit) == 0) ) continue;
    index = (int32)floor( (tData[irow]-useRange[0])/_binSize );
    if( index >= 0 && index < (int32)nRows ){
      if( zcol ){
	rateDat[index] += zData[irow];
	if(zerrcol)
	  rateErrDat[index] += zErrData[irow]*zErrData[irow];
	else
	  rateErrDat[index] += zData[irow]*zData[irow];
      }
      cntDat[index]++;
    }
  }

  // Tidy up

  if( zcol ){
    for(unsigned long irow=0; irow<outTable->rows(); irow++){
      rateDat[irow] /= _binSize;
      rateErrDat[irow] = sqrt(rateErrDat[irow])/_binSize;
    }

    delete cRateDat;
    delete cErrRateDat;
    delete cZDatCol;
    if(zerrcol) delete cZDatErrCol;
  }

  if( ratecol ){
    for(unsigned long irow=0; irow<outTable->rows(); irow++){
      rateDat[irow] = cntDat[irow]/_binSize;
      rateErrDat[irow] = std::sqrt(static_cast<double>(cntDat[irow]))/_binSize;
    }

    delete cRateDat;
    delete cErrRateDat;
  }

  delete cCntDat;
  delete cTDatCol;

//  // Delete empty rows if requested
//
//  if( booleanParameter("makesparserates") ){
//    std::vector<unsigned long> noZeroes;
//    for(unsigned long irow=0; irow<nRows; irow++)
//      if( cntDat[irow]!=0 ) noZeroes.push_back(irow);
//
//    unsigned long toRow = 0;
//    unsigned long ibegin=0;
//    for(unsigned long i=0; i<noZeroes.size(); i++){
//      unsigned long mRows;
//      if( noZeroes[i] == noZeroes.size()-1 || noZeroes[i]+1 != noZeroes[i+1] ){
//	mRows = i-ibegin+1;
//	outTable->copyRows(noZeroes[ibegin], toRow, mRows);
//	toRow += mRows;
//	ibegin = i+1;
//      }
//    }
//
//    outTable->deleteRows(toRow, outTable->rows() - toRow);
//  }

  if( ratecol || zcol )
    outTable->deleteColumn("COUNTS");

  if(! _destruct) delete cFlagData;

  // Determine E_MIN and E_MAX keywords

  DataSubspace dss(_intable);
  std::string eName = stringParameter("energycolumn");
  Column * eCol = _intable->column(eName);
  bool haveEmin=false, haveEmax=false, badEnergy=false;
  real64 emin = 0., emax = 0.;

  if( dss.components() > 0 ){
    DScomp * comp0 = dss.component(0);
    for(unsigned ifilt=0; ifilt<comp0->filters(); ifilt++){
      DSfilter * filter = comp0->filter(ifilt);
      if( filter->name()==eName ){
	if( !dss.isGlobal(*filter) ){
	  badEnergy=true;
	  continue;
	}
	if( filter->type()!=DSS::RangeT ) badEnergy=true;
	else{
	  double lower, upper;
	  RangeBound upperB, lowerB;
	  std::list<Range<double> > rlist = filter->ranges();
	  std::list<Range<double> >::const_iterator irange;
	  for(irange=rlist.begin(); irange!=rlist.end(); irange++){
	    irange->values(lower, upper, lowerB, upperB);
	    if( lowerB!=Unbounded ){
	      if( !haveEmin || (haveEmin && lower>emin) ) emin = lower;
	      haveEmin = true;
	    }
	    if( upperB!=Unbounded ){
	      if( !haveEmax || (haveEmax && upper<emax) ) emax = upper;
	      haveEmax = true;
	    }
	  }
	}
      }
    }

    if( !badEnergy ){
      for(unsigned icomp=1; icomp<dss.components(); icomp++){
	DScomp * comp = dss.component(icomp);
	for(unsigned ifilt=0; ifilt<comp->filters(); ifilt++){
	  DSfilter * filter = comp->filter(ifilt);
	  if( filter->name()==eName && !dss.isGlobal(*filter) ){
	    badEnergy=true;
	    break;
	  }
	}
      }
    }

    if( badEnergy )
      errstr << "A DSS energy filter has been found which cannot be "
	     << "incorporated into the calculation of E_MIN/MAX values "
	     << "for a rate curve." << Sas::warning("RateEnergy");
  }

  if( eCol->hasAttribute("TLMIN") ){
    real64 tlmin = eCol->attribute("TLMIN")->asReal();
    if( !haveEmin || (haveEmin && tlmin>emin) ) emin = tlmin;
    haveEmin = true;
  }
  if( eCol->hasAttribute("TLMAX") ){
    real64 tlmax = eCol->attribute("TLMAX")->asReal();
    if( !haveEmax || (haveEmax && tlmax<emax) ) emax = tlmax;
    haveEmax = true;
  }

  if( !haveEmax || !haveEmin ){
    real64 minRange, maxRange;
    getRange(eCol, minRange, maxRange);
    if( !haveEmin ) emin = minRange;
    if( !haveEmax ) emax = maxRange;
  }

  *outTable->addAttribute( "CHANMIN", "Minimum value of the energy column") = emin;
  *outTable->addAttribute( "CHANMAX", "Maximum value of the energy column") = emax;
  *outTable->addAttribute( "CHANTYPE", "Name of energy column used") = eName;

  // Write data subspace info

  if (writeDss) outTable << dss;

  // Calculate exposure keyword

  _getExposure(outTable);

  // Done
}

